"""ParPyDTK2 Cython interface"""

from libcpp cimport bool
from libcpp.string cimport string as std_string
from libcpp.vector cimport vector as std_vector
from libcpp.memory cimport shared_ptr
from mpi4py cimport libmpi as cMPI


ctypedef cMPI.MPI_Comm c_comm_t


cdef extern from 'src/dtk2.hpp' namespace 'parpydtk2' nogil:

    cdef cppclass IMeshDB:
        IMeshDB(c_comm_t comm) except +
        int ranks()
        int rank()
        c_comm_t comm()
        void begin_create() except +
        void create_vset() except +
        void create_vertices(int nv, const double *coords) except +
        void extract_vertices(double *coords) except +
        void assign_gids(int nv, const int *gids) except +
        void extract_gids(int *gids) except +
        void finish_create(bool trivial_gid) except +
        int size()
        bool empty()
        bool has_empty()
        const std_vector[int] &_m2s()
        void get_bbox(double *v) except +
        void get_gbbox(double *v) except +
        void create_field(const std_string &field_name, int dim) except +
        bool has_field(const std_string &field_name)
        int field_dim(const std_string &field_name) except +
        void assign_field(
            const std_string &field_name,
            const double *values
        ) except +
        void _assign_1st(
            const std_string &field_name,
            const double *values
        ) except +
        void extract_field(
            const std_string &field_name,
            double *values
        ) except +
        void _extract_1st(
            const std_string &field_name,
            double *values
        ) except +


    cdef cppclass Mapper:
        Mapper(
            shared_ptr[IMeshDB] B,
            shared_ptr[IMeshDB] G,
            const std_string &version,
            const std_string &date,
            bool profiling
        ) except +
        int ranks()
        int rank()
        c_comm_t comm()
        void set_dimension(int dim) except +
        void use_mmls()
        void use_spline()
        void use_n2n(bool matching)
        void set_basis(int basis) except +
        void use_knn_b(int knn) except +
        void use_knn_g(int knn) except +
        void use_radius_b(double r) except +
        void use_radius_g(double r) except +
        int check_method()
        int check_basis()
        int knn_b()
        int knn_g()
        double radius_b()
        double radius_g()
        int dimension()
        shared_ptr[IMeshDB] blue_mesh()
        shared_ptr[IMeshDB] green_mesh()
        void begin_initialization() except +
        void register_coupling_fields(
            const std_string &bf,
            const std_string &gf,
            bool direct
        ) except +
        bool has_coupling_fields(
            const std_string &bf,
            const std_string &gf,
            bool direct
        )
        void end_initialization() except +
        void begin_transfer()
        void transfer_data(
            const std_string &bf,
            const std_string &gf,
            bool direct
        ) except +
        void end_transfer()



#!python
#cython: language_level=3
#cython: boundscheck=False
#cython: embedsignature=True
#cython: wraparound=False

"""Interface mesh database

This module defines the representation of an interface mesh databse patch. It
is built on top of MOAB parallel mesh database. Since we are only interested in
the meshless methods, only point clouds are needed in :class:`IMeshDB`.

.. moduleauthor:: Qiao Chen <benechiao@gmail.com>
"""

cimport numpy as cnp
from cython.operator cimport dereference as deref
from libcpp cimport bool
from libcpp.string cimport string as std_string
from libcpp.vector cimport vector as std_vector
from libcpp.memory cimport make_shared
cimport parpydtk2 as dtk
from mpi4py cimport libmpi as cMPI
from mpi4py cimport MPI as pyMPI


ctypedef pyMPI.Comm comm_t
ctypedef cMPI.MPI_Comm c_comm_t


cdef extern from 'mpi-compat.hpp':
    pass


import numpy as np
from ._version import __version__


__version__ = __version__
__author__ = 'Qiao Chen'
__copyright__ = 'Copyright 2018, Qiao Chen'


cdef class IMeshDB(object):
    """Interface mesh database

    ParPyDTK2 utilizes MOAB as the underlying mesh database. MOAB is an array
    based mesh library that is adapted by DTK2. With array based mesh library,
    the memory usage and computational cost are lower than typical pointer
    based data structure. The mesh concept in this work is simple since only
    meshless methods are ultilized, the only additional attribute one needs
    is the `global IDs/handles`, which are used by both MOAB and DTK2. For most
    applications, the global IDs can be computed offline.

    One thing is not directly supported by IMeshDB is I/O. However, since this
    is a Python module and only points clouds are needed, one can easily uses
    a tool (e.g. `meshio <https://github.com/nschloe/meshio>`_) to load the
    mesh.

    Attributes
    ----------
    comm : MPI.Comm
        MPI communicator
    ranks : int
        size of comm
    rank : int
        rank of comm
    size : int
        point cloud size, i.e. number of vertices
    bbox : np.ndarray
        local bounding box array of shape (2,3)
    gbbox : np.ndarray
        global bounding box array of shape (2,3)
    """

    def __init__(self, comm=None):
        """Constructor
        
        Parameters
        ----------
        comm : MPI.Comm (optional)
            if no communicator or ``None`` is passed in, then ``MPI_COMM_WORLD``
            will be used

        Examples
        --------
        >>> # implicit communciator
        >>> import parpydtk2 as dtk
        >>> mdb = dtk.IMeshDB()

        >>> # explicit communicator
        >>> from mpi4py import MPI
        >>> import parpydtk2 as dtk
        >>> mdb = dtk.IMeshDB(MPI.COMM_WOLRD)
        """
        pass

    def __cinit__(self, comm_t comm=None):
        cdef c_comm_t comm_
        if comm is None:
            comm_ = cMPI.MPI_COMM_WORLD
        else:
            comm_ = comm.ob_mpi
        self.mdb = make_shared[dtk.IMeshDB](comm_)

    def __dealloc__(self):
        self.mdb.reset()

    @property
    def comm(self):
        """MPI.Comm: communicator"""
        cdef comm_t comm = pyMPI.Comm()
        comm.ob_mpi = self.mdb.get().comm()
        return comm

    @property
    def ranks(self):
        """int: Get the communicator size"""
        return self.mdb.get().ranks()

    @property
    def rank(self):
        """int: get the rank"""
        return self.mdb.get().rank()

    def begin_create(self):
        """Begin to create/manupilate the mesh

        This function must be called in order to let the mesh databse be
        aware that you will create meshes.

        See Also
        --------
        :func:`finish_create` : finish creating mesh
        """
        self.mdb.get().begin_create()

    def create_vertices(self, double[:, ::1] coords not None):
        """Create a set of coordinates

        .. note:: The ``coords`` must be C-ordering with ndim=2!

        Parameters
        ----------
        coords : np.ndarray
            nx3 coordinates in double precision

        See Also
        --------
        :func:`assign_gids` : assign global IDs
        :func:`extract_vertices`: extract vertex coordinates

        Examples
        --------
        >>> from parpydtk2 import *
        >>> import numpy as np
        >>> mdb1 = IMeshDB()
        >>> mdb1.begin_create()
        >>> verts = np.zeros((2,3))  # two nodes
        >>> verts[1][0] = 1.0
        >>> mdb1.create_vertices(verts)
        """
        assert coords.shape[1] == 3
        self.mdb.get().create_vertices(
            <int> coords.shape[0],
            <const double *> &coords[0, 0]
        )

    def extract_vertices(self):
        """Extract coordinate
        
        .. warning::
        
            This function should be called once you have finished
            :func:`create_vertices`.

        Returns
        -------
        np.ndarray:
            (nx3) array that stores the coordinate values

        See Also
        --------
        :attr:`size` : get the mesh size
        """
        cdef cnp.ndarray[double, ndim=2] co = np.empty((self.mdb.get().size(), 3))
        self.mdb.get().extract_vertices(<double *> co.data)
        return co

    def assign_gids(self, int[::1] gids):
        """Assign global IDs

        Internally, both DTK and MOAB use so-called global IDs/handles
        communications. Each node has its own local IDs/handles and a unique
        global ID.

        Parameters
        ----------
        gids : np.ndarray
            global IDs

        See Also
        --------
        :func:`create_vertices` : create vertices
        :func:`extract_gids` : extract global IDs
        """
        self.mdb.get().assign_gids(<int> gids.size, <const int *> &gids[0])

    def extract_gids(self):
        """Extract global IDs/handles
        
        .. warning::
        
            This function should be called once you have finished
            :func:`assign_gids`.

        Returns
        -------
        np.ndarray:
            array of :attr:`size` that stores the integer IDs
        """
        cdef cnp.ndarray[int, ndim=1] gids = \
            np.empty(self.mdb.get().size(), dtype='intc')
        self.mdb.get().extract_gids(<int *> gids.data)
        return gids

    def finish_create(self, trivial_gid=True):
        """finish mesh creation

        This method finalizes the interface mesh database by communicating
        the bounding boxes and empty partitions. Also, setting up the DTK
        managers happens here.

        .. warning::
        
            You must call this function once you have done with manupilating
            the mesh, i.e. vertices and global IDs.

        Parameters
        ----------
        trivial_gid : bool
            `True` if we use MOAB trivial global ID computation

        Notes
        -----
        By ``trivial_gid``, it means simply assigning the global IDs based on
        the size of the mesh. This is useful in serial settings or transferring
        solutions from a serial solver to a partitioned one.
        """
        cdef bool tg = <bool> 1 if trivial_gid else <bool> 0
        self.mdb.get().finish_create(tg)

    @property
    def size(self):
        """int: Get the size of a set"""
        return self.mdb.get().size()

    def empty(self):
        """Check if this is an empty partition"""
        return self.mdb.get().empty()

    def has_empty(self):
        """Check if an empty partition exists"""
        return self.mdb.get().has_empty()

    @property
    def bbox(self):
        """np.ndarray: local bounding box

        The bounding box is stored simply in a 2x3 array, where the first row
        stores the maximum bounds while minimum bounds for the second row.

        .. warning::

            Bounding box is valid only after :func:`finish_create`.
        
        See Also
        --------
        :attr:`gbbox`: global bounding box
        """
        cdef cnp.ndarray[double, ndim=2, mode='c'] box = \
            np.empty(shape=(2, 3), dtype='double')
        self.mdb.get().get_bbox(<double *> box.data)
        return box

    @property
    def gbbox(self):
        """np.ndarray: global bounding box
        
        The bounding box is stored simply in a 2x3 array, where the first row
        stores the maximum bounds while minimum bounds for the second row.

        .. warning::

            Bounding box is valid only after :func:`finish_create`.

        See Also
        --------
        :attr:`bbox`: local bounding box
        """
        cdef cnp.ndarray[double, ndim=2, mode='c'] box = \
            np.empty(shape=(2, 3), dtype='double')
        self.mdb.get().get_gbbox(<double *> box.data)
        return box

    def create_field(self, str field_name, int dim=1):
        """Create a data field for solution transfer

        This is the core function to register a field so that you can then
        transfer its values to other domains. The ``dim`` parameter determines
        the data type of the field. By default, it's 1, i.e. scalar fields.
        For each node, a tensor of (1x``dim``) can be registered. For instance,
        to transfer forces and displacements in FSI applications, ``dim`` is
        3 (for 3D problems).

        Parameters
        ----------
        field_name : str
            name of the field
        dim : int
            dimension of the field, i.e. scalar, vector, tensor

        Examples
        --------
        >>> from parpydtk2 import *
        >>> mdb1 = IMeshDB()
        >>> mdb1.begin_create()
        >>> mdb1.create_field('heat flux')
        """
        cdef std_string fn = <std_string> field_name.encode('UTF-8')
        self.mdb.get().create_field(fn, dim)

    def has_field(self, str field_name):
        """Check if a field exists

        Parameters
        ----------
        field_name : str
            name of the field

        Returns
        -------
        bool
            `True` if this meshdb has `field_name`
        """
        cdef std_string fn = <std_string> field_name.encode('UTF-8')
        return self.mdb.get().has_field(fn)

    def field_dim(self, str field_name):
        """Check the field data dimension

        Parameters
        ----------
        field_name : str
            name of the field

        Returns
        -------
        int
            data field dimension of `field_name`
        """
        cdef std_string fn = <std_string> field_name.encode('UTF-8')
        return self.mdb.get().field_dim(fn)

    def assign_field(self, str field_name, cnp.ndarray values not None):
        """Assign values to a field

        .. note:: `values` size must be at least size*dim

        Parameters
        ----------
        field_name : str
            name of the field
        values : np.ndarray
            input source values

        See Also
        --------
        :func:`extract_field` : extract value from a field
        :attr:`size` : check the size of a mesh set
        """
        cdef:
            std_string fn = <std_string> field_name.encode('UTF-8')
            int dim = self.mdb.get().field_dim(fn)
            int n = self.mdb.get().size() * dim
        assert values.size >= n
        self.mdb.get().assign_field(fn, <const double *> values.data)

    def resolve_empty_partitions(self, str field_name):
        """Resolve asynchronous values on empty partitions

        ParPyDTK2 doesn't expect :func:`assign_field` should be called
        collectively. Therefore, a collective call must be made for resolving
        empty partitions.

        .. warning:: This must be called collectively even on empty partitions

        .. note:: You should call this function following assignment

        Parameters
        ----------
        field_name : str
            name of the field

        Examples
        --------
        >>> if rank == 0:
        ...     mdb.assign_field('flux', values)
        >>> mdb.resolve_empty_partitions('flux')

        Notes
        -----
        This API is not available in C++ level, therefore, one needs to
        implement this if he/she wants to use the C++ API.
        """
        cdef:
            std_string fn = <std_string> field_name.encode('UTF-8')
            int dim = self.mdb.get().field_dim(fn)
            const std_vector[int] *m2s
            int i
            cnp.ndarray[double, ndim=1] buf
        if self.mdb.get().has_empty():
            buf = np.empty(dim)
            if self.mdb.get().rank() == 0:
                m2s = &self.mdb.get()._m2s()
                self.mdb.get()._extract_1st(fn, <double *> buf.data)
                for i in range(deref(m2s).size()):
                    self.comm.Isend(buf, dest=deref(m2s)[i])
            elif self.mdb.get().empty():
                req = self.comm.Irecv(buf, source=0)
                req.Wait()
                self.mdb.get()._assign_1st(fn, <const double *> buf.data)


    def extract_field(self, str field_name,
        double[::1] buffer=None, reshape=False):
        """Extact the values from a field

        .. warning:: if `buffer` is passed in, it must be 1D

        Parameters
        ----------
        field_name : str
            name of the field
        buffer : np.ndarray
            1D buffer
        reshape : bool
            `True` if we reshape the output, only for vectors/tensors

        Returns
        -------
        np.ndarray
            field data values
        """
        cdef:
            std_string fn = <std_string> field_name.encode('UTF-8')
            int n = self.mdb.get().size()
            int dim = self.mdb.get().field_dim(fn)
            cnp.ndarray[double, ndim=1] values
        if buffer is not None:
            assert len(buffer) >= n * dim
            values = np.asarray(buffer)
        else:
            values = np.empty(n * dim, dtype='double')
        self.mdb.get().extract_field(fn, <double *> values.data)
        if reshape and dim > 1:
            return values.reshape((n, dim))
        return values



#!python
#cython: language_level=3
#cython: boundscheck=False
#cython: embedsignature=True
#cython: wraparound=False

"""Interface solution transfer/mapper

This module wraps DTK2 meshless solution transfer operators, which can provide
convenient data remapping services.

.. moduleauthor:: Qiao Chen <benechiao@gmail.com>
"""

cimport numpy as cnp
from libcpp cimport bool
from libcpp.string cimport string as std_string
cimport parpydtk2 as dtk
from mpi4py cimport MPI as pyMPI
from .imeshdb cimport IMeshDB


ctypedef pyMPI.Comm comm_t


cdef extern from 'mpi-compat.hpp':
    pass


import numpy as np
import datetime
from ._version import __version__


__version__ = __version__
__author__ = 'Qiao Chen'
__copyright__ = 'Copyright 2018, Qiao Chen'


cdef class Mapper(object):
    """DTK2 wrapper

    The meshless methods in DTK2, including `modified moving least square`,
    `spline interpolation` and `nearest node projection` methods are wrapped
    within this class. The most advanced method is the MMLS fitting, which
    is my personal recommendataion.

    Attributes
    ----------
    blue_mesh : :py:class:`~parpydtk2.IMeshDB`
        blue mesh participant
    green_mesh : :py:class:`~parpydtk2.IMeshDB`
        green mesh participant
    comm : MPI.Comm
        MPI communicator
    ranks : int
        size of comm
    rank : int
        rank of comm
    dimension : int
        spatial dimension
    method : int
        method flag, either MMLS, SPLINE, or N2N
    basis : int
        flag of basis function for weighting schemes used by MMLS and SPLINE
    knn_b : int
        k-nearest neighborhood for searching on blue_mesh
    knn_g : int
        k-nearest neighborhood for searching on green_mesh
    radius_b : float
        radius used for searching on blue_mesh
    radius_g : float
        radius used for searching on green_mesh
    """

    def __init__(self, blue, green, profiling=True):
        """Constructor

        Parameters
        ----------
        blue_mesh : :py:class:`~parpydtk2.IMeshDB`
            blue mesh participant
        green_mesh : :py:class:`~parpydtk2.IMeshDB`
            green mesh participant
        profiling : bool (optional)
            whether or not do timing report, default is ``True``.

        Examples
        --------
        >>> from parpydtk2 import *
        >>> blue, green = IMeshDB(), IMeshDB()
        >>> # initialize blue and green
        >>> mapper = Mapper(blue=blue,green=green)
        >>> # do work with mapper
        """
        pass
    
    def __cinit__(self, IMeshDB blue, IMeshDB green, profiling=True):
        cdef:
            std_string version = __version__.encode('UTF-8')
            std_string date = \
                datetime.datetime.now().strftime('%b %d %Y %H:%M:%S').encode('UTF-8')
            bool prof = <bool> 1 if profiling else <bool> 0
        self.mp = new dtk.Mapper(blue.mdb, green.mdb, version, date, prof)
        self.blue = blue
        self.green = green
    
    def __dealloc__(self):
        del self.mp

    @property
    def blue_mesh(self):
        """:py:class:`~parpydtk2.IMeshDB`: blue mesh"""
        return self.blue

    @property
    def green_mesh(self):
        """:py:class:`~parpydtk2.IMeshDB`: green mesh"""
        return self.green

    @property
    def ranks(self):
        """int: Check the total process number

        See Also
        --------
        :attr:`rank` : "my" rank
        """
        return self.mp.ranks()

    @property
    def rank(self):
        """int: Check "my" rank

        See Also
        --------
        :attr:`ranks` : get the total communicator size
        :attr:`comm` : MPI communicator
        """
        return self.mp.rank()

    @property
    def comm(self):
        """MPI.Comm: Get the communicator

        See Also
        --------
        :attr:`ranks` : get the total communicator size
        :attr:`rank` : "my" rank
        """
        cdef comm_t comm = pyMPI.Comm()
        comm.ob_mpi = self.mp.comm()
        return comm

    @property
    def dimension(self):
        """int: Get the problem dimension

        .. note:: this is the spacial dimension
        """
        return self.mp.dimension()

    @dimension.setter
    def dimension(self, int dim):
        self.mp.set_dimension(dim)

    @property
    def method(self):
        """int: Get the method tag

        See Also
        --------
        :attr:`basis` : the basis function and order attribute
        """
        return self.mp.check_method()

    @method.setter
    def method(self, int m):
        if m == 0:
            # mls
            self.mp.use_mmls()
        elif m == 1:
            self.mp.use_spline()
        elif m == 2:
            self.mp.use_n2n(<bool> 0)
        else:
            raise AttributeError('unknown method')

    def set_matching_flag_n2n(self, bool matching):
        """Set the matching flag for N2N

        .. note:: this function will not throw even if you dont use n2n

        Parameters
        ----------
        matching : bool
            `True` if the interfaces are matching
        """
        cdef bool flag = <bool> 1 if matching else <bool> 0
        self.mp.use_n2n(flag)

    @property
    def basis(self):
        """int: Get the basis function flag

        See Also
        --------
        :attr:`method` : get the method tag
        """
        return self.mp.check_basis()

    @basis.setter
    def basis(self, int bf):
        self.mp.set_basis(bf)

    @property
    def knn_b(self):
        """int: KNN of blue mesh

        .. note:: if blue does not use KNN, then -1 returned

        See Also
        --------
        :attr:`knn_g` : green knn
        """
        return self.mp.knn_b()
    
    @knn_b.setter
    def knn_b(self, int knn):
        self.mp.use_knn_b(knn)

    @property
    def knn_g(self):
        """int: KNN of green mesh

        .. note:: if green does not use KNN, then -1 returned

        See Also
        --------
        :attr:`knn_b` : blue knn
        """
        return self.mp.knn_g()

    @knn_g.setter
    def knn_g(self, int knn):
        self.mp.use_knn_g(knn)
    
    @property
    def radius_b(self):
        """float: physical domain radius support for blue mesh

        .. note:: if blue does not use RBF-search, then -1.0 returned

        See Also
        --------
        :attr:`radius_g` : green radius
        """
        return self.mp.radius_b()

    @radius_b.setter
    def radius_b(self, double r):
        self.mp.use_radius_b(r)

    @property
    def radius_g(self):
        """float: physical domain radius support for green mesh

        .. note:: if green does not use RBF-search, then -1.0 returned

        See Also
        --------
        :attr:`radius_b` : blue radius
        """
        return self.mp.radius_g()

    @radius_g.setter
    def radius_g(self, double r):
        self.mp.use_radius_g(r)

    def begin_initialization(self):
        """Initialization starter

        This is a must-call function in order to indicate mapper that you are
        about to initialize/register coupling fields

        See Also
        --------
        :func:`register_coupling_fields` : register coupled fields
        :func:`end_initialization` : finish initialization
        """
        self.mp.begin_initialization()

    def register_coupling_fields(self, str bf, str gf, bool direct):
        """register a coupled fields

        .. note:: we use boolean to indicate direction

        Parameters
        ----------
        bf : str
            blue mesh field name
        gf : str
            green mesh field name
        direct : bool
            `True` for blue->green, `False` for the opposite

        See Also
        --------
        :func:`transfer_data` : transfer a coupled data fields
        """
        cdef:
            std_string bf_ = <std_string> bf.encode('UTF-8')
            std_string gf_ = <std_string> gf.encode('UTF-8')
            bool dr = <bool> 1 if direct else <bool> 0
        self.mp.register_coupling_fields(bf_, gf_, dr)

    def has_coupling_fields(self, str bf, str gf, bool direct):
        """Check if a coupled fields exists

        Returns
        -------
        bool
            ``True`` if (bf,gf) exists in the ``direct`` direction
        """
        cdef:
            std_string bf_ = <std_string> bf.encode('UTF-8')
            std_string gf_ = <std_string> gf.encode('UTF-8')
            bool dr = <bool> 1 if direct else <bool> 0
        return self.mp.has_coupling_fields(bf_, gf_, dr)

    def end_initialization(self):
        """Initialization closer

        This is a must-call function in order to tell the mapper we are ready

        See Also
        --------
        :func:`begin_initialization` : initialization starter
        """
        self.mp.end_initialization()

    def begin_transfer(self):
        """Transfer starter

        This is a must-call function to inidate the beginning of a transferring
        block

        See Also
        --------
        :func:`end_transfer` : transfer closer
        :func:`tranfer_data` : transfer a coupled data fields
        """
        self.mp.begin_transfer()

    def transfer_data(self, str bf, str gf, bool direct):
        """Transfer (bf, gf) in the ``direct`` direction

        Parameters
        ----------
        bf : str
            blue mesh field name
        gf : str
            green mesh field name
        direct : bool
            ``True`` for blue->green, ``False`` for the opposite

        See Also
        --------
        :func:`register_coupling_fields` : register coupled fields
        """
        cdef:
            std_string bf_ = <std_string> bf.encode('UTF-8')
            std_string gf_ = <std_string> gf.encode('UTF-8')
            bool dr = <bool> 1 if direct else <bool> 0
        # resolve empty partitions
        if dr:
            src = self.blue
            tag_src = bf
        else:
            src = self.green
            tag_src = gf
        self.mp.transfer_data(bf_, gf_, dr)

    def end_transfer(self):
        """Transfer closer

        This is a must-call function to indicate we have finished a sequence of
        transferring requests

        See Also
        --------
        :func:`begin_transfer` : transfer starter
        """
        self.mp.end_transfer()


"""Main module interface of ParPyDTK2

.. moduleauthor:: Qiao Chen <benechiao@gmail.com>

Attributes
----------
B2G : bool
    boolean flag of ``True`` denotes transferring direction from ``blue`` to
    ``green``
G2B : bool
    boolean flag of ``False`` denotes transferring direction from ``green`` to
    ``blue``
MMLS : int
    flag (0) represents using `modified moving least square` method
SPLINE : int
    flag (1) represents using `spline interpolation` method
N2N : int
    flag (2) represents using `nearest node projection` method
WENDLAND2 : int
    flag (0) represents using `Wendland 2nd-order` RBF weights
WENDLAND4 : int
    flag (1) represents using `Wendland 4th-order` RBF weights
WENDLAND6 : int
    flag (2) represents using `Wendland 6th-order` RBF weights
WU2 : int
    flag (3) represents using `Wu 2nd-order` RBF weights
WU4 : int
    flag (4) represents using `Wu 4th-order` RBF weights
WU6 : int
    flag (5) represents using `Wu 6th-order` RBF weights
BUHMANN3 : int
    flag (6) represents using `Buhmann 3rd-order` RBF weights
"""

import sys
import parpydtk2.error_handle as error
from ._version import __version__
from .imeshdb import IMeshDB
from .mapper import Mapper

__b2g__ = True
__g2b__ = False
B2G = __b2g__
G2B = __g2b__

__mmls__ = 0
__spline__ = 1
__n2n__ = 2
MMLS = __mmls__
SPLINE = __spline__
N2N = __n2n__

__wendland2__ = 0
__wendland4__ = 1
__wendland6__ = 2
WENDLAND2 = __wendland2__
WENDLAND4 = __wendland4__
WENDLAND6 = __wendland6__

__wu2__ = 3
__wu4__ = 4
__wu6__ = 5
WU2 = __wu2__
WU4 = __wu4__
WU6 = __wu6__

__buhmann3__ = 6
BUHMANN3 = __buhmann3__


def get_include():
    """Get the abs include path"""
    import os
    return os.path.dirname(os.path.abspath(__file__))



from distutils.command.clean import clean
import os
import sys
import re
import codecs
import numpy
from setuptools import setup, Extension, find_packages
try:
    import mpi4py
except ImportError:
    sys.stderr.write('You must install mpi4py first!\n')
    sys.stderr.flush()
    sys.exit(-1)

if sys.version_info[:2] < (3, 5):
    print('ParPyDTK2 requires Python 3.5 or higher.')
    sys.exit(-1)

vfile = open('parpydtk2/_version.py', mode='r')
vstr_raw = vfile.read()
vstr_find = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]", vstr_raw, re.M)
if vstr_find:
    version = vstr_find.group(1)
else:
    raise RuntimeError('Unable to find __version__ in parpydtk2/_version.py.')
vfile.close()

classifiers = [
    'Programming Language :: Python :: 3.5',
    'Programming Language :: Python :: 3.6',
    'Programming Language :: C++',
    'Topic :: Scientific/Engineering',
    'Topic :: Software Development',
    'Operating System :: POSIX :: Linux',
    'Intended Audience :: Science/Research',
    'License :: OSI Approved :: MIT License',
]

install_requires = [
    'numpy',
    'setuptools',
]

package_data = {
    'parpydtk2': [
        '*.pxd',
        '*.pyx',
        '*.cpp',
        '*.hpp',
        os.path.join('src', '*.hpp'),
    ]
}

# set the compiler to mpi
os.environ['CC'] = mpi4py.get_config()['mpicxx']

LEN1, LEN2 = 80, 5


def ensure_cpp11():
    import tempfile
    from distutils.ccompiler import get_default_compiler, new_compiler
    import distutils
    cmp = new_compiler(get_default_compiler())
    tmp_source = os.path.join(tempfile.gettempdir(), 'foo.cpp')
    f = open(tmp_source, 'w')
    f.write('int main(){return 0;}')
    f.close()

    def safe_remove(filename):
        try:
            os.remove(filename)
        except OSError:
            pass
    try:
        cmp.compile([tmp_source], extra_postargs=['-std=c++11'])
        safe_remove(tmp_source)
        safe_remove(tempfile.gettempprefix())
    except distutils.errors.CompileError:
        safe_remove(tmp_source)
        safe_remove(tempfile.gettempprefix())
        sys.stderr.write('You must use a compiler that supports C++11\n')
        sys.stderr.flush()
        sys.exit(-1)


ensure_cpp11()


# configure paths
include_dirs = [
    '.',
    numpy.get_include(),
    mpi4py.get_include()
]

libraries = [
    'MOAB',
    'dtk_moabadapters',
    'dtk_interface',
    'teuchoscomm',
    'teuchosparameterlist',
    'teuchoscore',
    'dtk_operators'
]

library_dirs = None


def configure_lib(env):
    # check env
    root = os.environ.get(env, None)
    if root is not None:
        return ([os.path.join(root, 'include')],
                [os.path.join(root, 'lib')])
    return None


MOAB = configure_lib('PARPYDTK2_MOAB_ROOT')
DTK = configure_lib('PARPYDTK2_DTK_ROOT')


if MOAB is not None:
    if library_dirs is None:
        library_dirs = []
    include_dirs += MOAB[0]
    library_dirs += MOAB[1]
if DTK is not None:
    if library_dirs is None:
        library_dirs = []
    include_dirs += DTK[0]
    library_dirs += DTK[1]


if MOAB is None or DTK is None:
    def configure():
        # check arguments
        prefix = None
        if '--user' in sys.argv:
            from site import USER_BASE
            prefix = USER_BASE
        else:
            for arg in sys.argv[1:]:
                if arg.startswith('--home'):
                    prefix = arg.split('=')[1]
                    if prefix == '~':
                        prefix = os.environ.get('HOME')
                elif arg.startswith('--prefix'):
                    prefix = arg.split('=')[1]
        if prefix is not None:
            return ([os.path.join(prefix, 'include')],
                    [os.path.join(prefix, 'lib')])
        return None
    CONF = configure()
    if CONF is not None:
        if library_dirs is None:
            library_dirs = []
        include_dirs += CONF[0]
        library_dirs += CONF[1]


rpath = library_dirs

exts = [
    Extension(
        'parpydtk2.imeshdb',
        ['parpydtk2/imeshdb.cpp'],
        include_dirs=include_dirs,
        language='c++',
        libraries=libraries,
        library_dirs=library_dirs,
        runtime_library_dirs=rpath,
        extra_compile_args=['-std=c++11'],
    ),
    Extension(
        'parpydtk2.mapper',
        ['parpydtk2/mapper.cpp'],
        include_dirs=include_dirs,
        language='c++',
        libraries=libraries,
        library_dirs=library_dirs,
        runtime_library_dirs=rpath,
        define_macros=[('LEN1', LEN1), ('LEN2', LEN2)],
        extra_compile_args=['-std=c++11'],
    )
]


class MyClean(clean):
    def run(self):
        import subprocess as sp
        super().run()
        sp.Popen(
            'rm -rf parpydtk2/__pycache__ __pycache__ dist parpydtk2.egg-info',
            executable='/bin/bash',
            shell=True
        )
        if self.all:
            sp.Popen(
                'rm -rf parpydtk2/*.so',
                executable='/bin/bash',
                shell=True
            )


setup(
    name='parpydtk2',
    version=version,
    description='Python Wrapper for Parallel DTK2+MOAB',
    author='Qiao Chen',
    author_email='benechiao@gmail.com',
    keywords='Math',
    long_description=codecs.open('README.rst', encoding='utf-8').read(),
    url='https://github.com/chiao45/parpydtk2',
    packages=find_packages(),
    license='MIT',
    ext_modules=exts,
    classifiers=classifiers,
    cmdclass={'clean': MyClean},
    install_requires=install_requires,
    package_data=package_data,
)



"""The error handler module

.. moduleauthor:: Qiao Chen <benechiao@gmail.com>

Attributes
----------
ERROR_CODE : int
    set this to nonzero values one exceptions have been raised

Examples
--------
>>> import parpydtk2 as dtk
>>> try:
...     # your programs here
... except Expection:
...     dtk.error.ERROR_CODE = 1
...     raise
"""

import atexit


ERROR_CODE = 0


def _atexit():
    if ERROR_CODE != 0:
        import traceback
        import sys
        from mpi4py import MPI

        traceback.print_exc()
        sys.stderr.flush()
        if MPI.Is_initialized() and MPI.COMM_WORLD.size > 1:
            MPI.COMM_WORLD.Abort(ERROR_CODE)


atexit.register(_atexit)
